/****************************************************************************
** Dooble - The Secure Internet Web Browser
**
** Copyright (c) 2008, 2009, 2010, 2011, 2012 Alexis Megas,
** Gunther van Dooble, and the Dooble Team.
** All rights reserved.
**
** License: GPL2 only:
** This program is free software; you can redistribute it and/or modify
** it under the terms of the GNU General Public License as published by
** the Free Software Foundation; version 2 of the License only.
**
** This program is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
** GNU General Public License for more details.
**
** You should have received a copy of the GNU General Public License
** along with this program; if not, write to the Free Software
** Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
** or see here: http://www.gnu.org/licenses/gpl.html
**
** For the WebKit library, please see: http://webkit.org.
**
** THE CODE IS PROVIDED BY THE AUTHORS ''AS IS'' AND ANY
** EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
** IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
** PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR
** ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
** DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE
** GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
** INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER
** IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
** ARISING IN ANY WAY OUT OF THE USE OF THIS APPLICATION, EVEN IF ADVISED
** OF THE POSSIBILITY OF SUCH DAMAGE.
**
** Please report all praise, requests, bugs, and problems to the project
** team and administrators: http://sf.net/projects/dooble.
**
** You can find us listed at our project page. New team members are welcome.
** The name of the authors should not be used to endorse or promote products
** derived from Dooble without specific prior written permission.
** If you use this code for other projects, please let us know.
**
** Web sites:
**   http://sf.net/projects/dooble
**   http://dooble.sf.net
****************************************************************************/

#include <QSettings>
#include <QSqlQuery>
#include <QSqlDatabase>
#include <QFileIconProvider>

#include "dmisc.h"
#include "dooble.h"
#include "dbookmarkspopup.h"
#include "dbookmarkswindow.h"

dbookmarkspopup::dbookmarkspopup(void):QWidget()
{
  m_folderOid = -1;
  ui.setupUi(this);
#ifdef Q_WS_MAC
  setAttribute(Qt::WA_MacMetalStyle, false);
#endif
  ui.folders->setModel(dooble::s_bookmarksFolderModel);
  connect(ui.folders,
	  SIGNAL(itemSelected(const QModelIndex &)),
	  this,
	  SLOT(slotFolderSelected(const QModelIndex &)));
  connect(ui.titleLineEdit,
	  SIGNAL(returnPressed(void)),
	  this,
	  SLOT(slotTitleChanged(void)));
  connect(ui.closePushButton,
	  SIGNAL(clicked(void)),
	  this,
	  SLOT(hide(void)));
  connect(ui.addFolderPushButton,
	  SIGNAL(clicked(void)),
	  this,
	  SLOT(slotAddFolder(void)));
  connect(ui.descriptionTextEdit,
	  SIGNAL(textChanged(void)),
	  this,
	  SLOT(slotDescriptionChanged(void)));
  connect(ui.deleteBookmarkPushButton,
	  SIGNAL(clicked(void)),
	  this,
	  SLOT(slotDeleteBookmark(void)));
  connect(dooble::s_bookmarksFolderModel,
	  SIGNAL(dataChanged(const QModelIndex &,
			     const QModelIndex &)),
	  this,
	  SLOT(slotFolderDataChanged(const QModelIndex &,
				     const QModelIndex &)));
  setWindowFlags(Qt::Popup);
  slotSetIcons();
}

dbookmarkspopup::~dbookmarkspopup()
{
}

void dbookmarkspopup::exec(const QPoint &point, const QUrl &url)
{
  m_url = url;

  /*
  ** Retrieve the bookmark's folder, title, and description.
  */

  {
    QSqlDatabase db = QSqlDatabase::addDatabase("QSQLITE", "bookmarks");

    db.setDatabaseName(dooble::s_homePath +
		       QDir::separator() + "bookmarks.db");

    if(db.open())
      {
	QString hashType(dooble::s_settings.value("settingsWindow/hashType").
			 toString());
	QString queryStr("");
	QSqlQuery query(db);

	queryStr = QString("SELECT title, description, folder_oid "
			   "FROM bookmarks WHERE "
			   "url_hash = '%1' AND "
			   "passphrase_hash = '%2'").
	  arg(dmisc::hash(url.toEncoded(QUrl::StripTrailingSlash)).toHex().
	      constData()).
	  arg(dmisc::passphraseHash(hashType));

	if(query.exec(queryStr))
	  if(query.next())
	    {
	      m_folderOid = query.value(2).toInt();
	      m_title =
		QString::fromUtf8
		(dmisc::decodedString
		 (QByteArray::fromBase64
		  (query.value(0).toByteArray())).constData());

	      QString description
		(dmisc::decodedString
		 (QByteArray::fromBase64
		  (query.value(1).toByteArray())).constData());

	      ui.titleLineEdit->setText(m_title);
	      ui.descriptionTextEdit->blockSignals(true);
	      ui.descriptionTextEdit->setPlainText(description);
	      ui.descriptionTextEdit->blockSignals(false);
	    }
      }

    db.close();
  }

  QSqlDatabase::removeDatabase("bookmarks");
  ui.titleLineEdit->selectAll();
  ui.titleLineEdit->setFocus();

  if(dooble::s_bookmarksFolderModel->index(0, 0).isValid())
    {
      /*
      ** Recursively search for the bookmark's folder.
      ** If found, select the folder and expand its parents.
      */

      QModelIndexList list(dooble::s_bookmarksFolderModel->
			   match(dooble::s_bookmarksFolderModel->index(0, 0),
				 Qt::UserRole + 2,
				 QVariant(m_folderOid),
				 1,
				 Qt::MatchRecursive));

      if(!list.isEmpty())
	{
	  ui.folders->setExpanded(list.first(), true);
	  ui.folders->selectionModel()->select
	    (list.first(), QItemSelectionModel::ClearAndSelect);
	  ui.folders->selectionModel()->setCurrentIndex
	    (list.first(), QItemSelectionModel::Current);
	}

      list.clear();
    }

  move(point);
  show();
}

void dbookmarkspopup::populate(void)
{
  /*
  ** We don't do much here.
  */
}

void dbookmarkspopup::updateBookmark(void)
{
  {
    QSqlDatabase db = QSqlDatabase::addDatabase("QSQLITE",
						"bookmarks");

    db.setDatabaseName(dooble::s_homePath +
		       QDir::separator() + "bookmarks.db");

    if(db.open())
      {
	QString hashType
	  (dooble::s_settings.value("settingsWindow/hashType").
	   toString());
	QSqlQuery query(db);

	query.exec("PRAGMA synchronous = OFF");
	query.prepare
	  ("UPDATE bookmarks SET title = ?, "
	   "description = ? "
	   "WHERE url_hash = ? AND "
	   "folder_oid = ? AND "
	   "passphrase_hash = ?");
	query.bindValue
	  (0,
	   dmisc::encodedString(m_title.toUtf8(), true).toBase64());
	query.bindValue
	  (1,
	   dmisc::encodedString(ui.descriptionTextEdit->toPlainText().
				trimmed().toUtf8(), true).toBase64());
	query.bindValue
	  (2, dmisc::hash(m_url.toEncoded(QUrl::StripTrailingSlash)).
	   toHex().constData());
	query.bindValue
	  (3, m_folderOid);
	query.bindValue
	  (4, dmisc::passphraseHash(hashType));
	query.exec();
      }

    db.close();
  }

  QSqlDatabase::removeDatabase("bookmarks");
  emit changed();
}

void dbookmarkspopup::slotSetIcons(void)
{
  QPixmap pixmap;
  QSettings settings
    (dooble::s_settings.value("iconSet").toString(), QSettings::IniFormat);

  pixmap.load(settings.value("bookmarksPopup/bookmarkIcon").toString());
  ui.icon->setPixmap(pixmap.scaled(QSize(32, 32),
				   Qt::KeepAspectRatio,
				   Qt::SmoothTransformation));
  ui.closePushButton->setIcon
    (QIcon(settings.value("bookmarksPopup/closeButtonIcon").toString()));
  ui.addFolderPushButton->setIcon
    (QIcon(settings.value("bookmarksPopup/addButtonIcon").toString()));
  emit iconsChanged();
}

void dbookmarkspopup::slotAddFolder(void)
{
  QItemSelection selection(ui.folders->selectionModel()->selection());

  if(dooble::s_bookmarksWindow)
    dooble::s_bookmarksWindow->addFolder();

  if(selection.indexes().value(0).isValid())
    ui.folders->selectionModel()->select
      (selection, QItemSelectionModel::ClearAndSelect);
  else
    ui.folders->selectionModel()->select
      (dooble::s_bookmarksFolderModel->index(0, 0),
       QItemSelectionModel::ClearAndSelect);
}

void dbookmarkspopup::slotTitleChanged(void)
{
  QString title(ui.titleLineEdit->text().trimmed());

  if(title.isEmpty())
    {
      ui.titleLineEdit->setText(m_title);
      ui.titleLineEdit->selectAll();
      return;
    }
  else
    {
      m_title = title;
      ui.titleLineEdit->selectAll();
    }

  updateBookmark();
}

void dbookmarkspopup::slotDeleteBookmark(void)
{
  {
    QSqlDatabase db = QSqlDatabase::addDatabase("QSQLITE", "bookmarks");

    db.setDatabaseName(dooble::s_homePath +
		       QDir::separator() + "bookmarks.db");

    if(db.open())
      {
	QString hashType
	  (dooble::s_settings.value("settingsWindow/hashType").
	   toString());
	QSqlQuery query(db);

	query.exec("PRAGMA synchronous = OFF");
	query.exec(QString("DELETE FROM bookmarks WHERE "
			   "folder_oid = %1 AND "
			   "url_hash = '%2' AND "
			   "passphrase_hash = '%3'").
		   arg(m_folderOid).
		   arg(dmisc::hash(m_url.
				   toEncoded(QUrl::StripTrailingSlash)).
		       toHex().constData()).
		   arg(dmisc::passphraseHash(hashType)));
      }

    db.close();
  }

  QSqlDatabase::removeDatabase("bookmarks");
  emit changed();
  close();
}

void dbookmarkspopup::slotFolderSelected(const QModelIndex &index)
{
  if(index.isValid())
    {
      QStandardItem *item = dooble::s_bookmarksFolderModel->itemFromIndex
	(index);

      if(item)
	{
	  m_folderOid = item->data(Qt::UserRole + 2).toInt();

	  {
	    QSqlDatabase db = QSqlDatabase::addDatabase("QSQLITE",
							"bookmarks");

	    db.setDatabaseName(dooble::s_homePath +
			       QDir::separator() + "bookmarks.db");

	    if(db.open())
	      {
		QString hashType
		  (dooble::s_settings.value("settingsWindow/hashType").
		   toString());
		QSqlQuery query(db);

		query.exec("PRAGMA synchronous = OFF");
		query.prepare
		  ("UPDATE bookmarks SET folder_oid = ? "
		   "WHERE url_hash = ? AND "
		   "passphrase_hash = ?");
		query.bindValue
		  (0, m_folderOid);
		query.bindValue
		  (1, dmisc::hash(m_url.toEncoded(QUrl::StripTrailingSlash)).
		   toHex().constData());
		query.bindValue
		  (2, dmisc::passphraseHash(hashType));
		query.exec();
	      }

	    db.close();
	  }

	  QSqlDatabase::removeDatabase("bookmarks");
	  emit changed();
	}
    }
}

void dbookmarkspopup::slotFolderDataChanged
(const QModelIndex &topLeft, const QModelIndex &bottomRight)
{
  if(!dooble::s_bookmarksWindow)
    return;

  Q_UNUSED(bottomRight);

  QModelIndexList list(ui.folders->selectionModel()->selectedIndexes());

  if(!list.isEmpty())
    if(topLeft == list.first())
      dooble::s_bookmarksWindow->renameFolder(topLeft);

  list.clear();
}

void dbookmarkspopup::slotDescriptionChanged(void)
{
  updateBookmark();
}
